%
% DA_critic_sim.m
%
% Simulate a model of a DA cell-driven critic.
%
% Last modification: 1/15/07


% Set simulation parameters.
pre_CS_iters = 40;
CS_iters = 40;
R_CS_offset_iters = 34;
R_iters = 5;
num_crit_trials = 7;
post_trial_iters = 10;
CS_on = [ 1 1 1 1 1 1 1 ];
R_on = [ 1 1 1 1 1 1 0 ];
% CS_on = [ 1 1 1 0 ];
% R_on = [ 1 1 0 0 ];
num_iters = num_crit_trials * (pre_CS_iters + CS_iters) + post_trial_iters;
                          % number of iterations in simulation (maximum)
iter_duration = 100.0;    % duration of each iteration (in ms)
monitor_interval = 100;   % increment of iterations at which we report how 
                          % much time has elapsed
init_unit_noise = 0.001;  % amplitude of initial unit activity noise
load_trained_wts = 0;
learning_on = 1;
save_trained_wts = 1;
LTP_LTD_postsyn_thresh = 0.7;
LTD_presyn_thresh = 0.3;
LTP_presyn_thresh = 0.7;
LTP_LTD_rew_thresh = 0.7;
LTD_decay_coeff = 0.5;
wt_decay_rate = 0.0001;

%
% Initialize the model.
%

% Load simulation run parameters.
load sim_run_params;

% Load model parameters (connectivity and cell/channel-types).
if (load_trained_wts == 1)
    load(train_model_fname);
else
    load(start_model_fname);
end;

% Load physiological parameters (cells and channels).
load(phys_fname);

% Run through all of the areas and extract the appropriate area indices.
for xx = 1:size(Area_name,1)
    the_area_name = sscanf(Area_name(xx,:),'%s');
    if (strcmp(the_area_name, 'CS_WM'))
        CS_WM_ind = Area_num(xx);
    elseif (strcmp(the_area_name, 'RewardIn'))
        RewardIn_ind = Area_num(xx);
    elseif (strcmp(the_area_name, 'TonicDAAff'))
        TonicDAAff_ind = Area_num(xx);        
    elseif (strcmp(the_area_name, 'RewardCueE'))
        RewardCueE_ind = Area_num(xx);        
    elseif (strcmp(the_area_name, 'RewardCueI'))
        RewardCueI_ind = Area_num(xx);  
    elseif (strcmp(the_area_name, 'RewardPred'))
        RewardPred_ind = Area_num(xx);  
    elseif (strcmp(the_area_name, 'DACell'))
        DACell_ind = Area_num(xx);           
    end;
end;

% Find the afferent input unit indices.
lo_bound = 1;
for xx = 1:CS_WM_ind - 1
    lo_bound = lo_bound + Area_xy(xx,3);
end;
hi_bound = lo_bound + Area_xy(CS_WM_ind,3) - 1;
CS_WMLoUnit_ind = lo_bound;
CS_WMHiUnit_ind = hi_bound;

lo_bound = 1;
for xx = 1:RewardIn_ind - 1
    lo_bound = lo_bound + Area_xy(xx,3);
end;
hi_bound = lo_bound + Area_xy(RewardIn_ind,3) - 1;
RewardInUnit_ind = lo_bound;

lo_bound = 1;
for xx = 1:TonicDAAff_ind - 1
    lo_bound = lo_bound + Area_xy(xx,3);
end;
hi_bound = lo_bound + Area_xy(TonicDAAff_ind,3) - 1;
TonicDAAffUnit_ind = lo_bound;

lo_bound = 1;
for xx = 1:RewardCueE_ind - 1
    lo_bound = lo_bound + Area_xy(xx,3);
end;
hi_bound = lo_bound + Area_xy(RewardCueE_ind,3) - 1;
RewardCueEUnit_ind = lo_bound;

lo_bound = 1;
for xx = 1:RewardPred_ind - 1
    lo_bound = lo_bound + Area_xy(xx,3);
end;
hi_bound = lo_bound + Area_xy(RewardPred_ind,3) - 1;
RewardPredLoUnit = lo_bound;
RewardPredHiUnit = hi_bound;

lo_bound = 1;
for xx = 1:DACell_ind - 1
    lo_bound = lo_bound + Area_xy(xx,3);
end;
hi_bound = lo_bound + Area_xy(DACell_ind,3) - 1;
DACellUnit_ind = lo_bound;

% Initialize the reward prediction apparatus.
num_CS = 4;
num_rew_delays = 40;
min_rew_delay = 10;   % in iterations
CS_activs = zeros(num_CS,1);
CS_past_activs = zeros(num_CS,1);
CS_thresh = 0.8;
rew_pred_out = zeros(num_rew_delays,1);
rew_pred_netin = zeros(num_rew_delays,1);
rew_pred_CW_ind = zeros(num_rew_delays,num_CS);
rew_pred_wts = zeros(num_rew_delays,num_CS);
rew_pred_ic_ct = zeros(num_rew_delays,num_CS);
rew_pred_act_wts = zeros(num_rew_delays,num_CS);
% rew_pred_learn_rate = 1.0;
rew_pred_learn_rate = 0.5;
rew_pred_maxwt = 1.0;
DA_rew_pred_wts = zeros(num_rew_delays,1);
tonic_DA_input = 0.2;

%
% Load the rew_pred_CW_ind indices
%

% Find the CS to predicted reward fascicle.
for xx = 1:size(Conn_name,1)
    the_conn_name = sscanf(Conn_name(xx,:),'%s');
    if (strcmp(the_conn_name, 'CS_WMtoRewPred'))
        CS_WMtoRewPred_ind = xx;
    end;
end;
    
% Find the index for the first connection of the desired fascicle.
first_conn_ind = 1;
for xx = 1:CS_WMtoRewPred_ind - 1
    first_conn_ind = first_conn_ind + NC(xx);
end;

% Get the connection to and from addresses and the weights for the desired 
% fascicle.
addr = double(CA(:,first_conn_ind:first_conn_ind + ...
    NC(CS_WMtoRewPred_ind) - 1));

% Subtract the unit offsets to get the relative connection addresses.
addr(1,:) = addr(1,:) - unit_offset(Conn_fromto(CS_WMtoRewPred_ind,2));
addr(2,:) = addr(2,:) - unit_offset(Conn_fromto(CS_WMtoRewPred_ind,1));

% For each reward prediction unit (with a different delay)...
for to_ind = 1:Area_xy(Conn_fromto(CS_WMtoRewPred_ind,2),3) 
    % For all connections to this unit, save the value of the address into
    % CA and CW so we can use it later.
    for xx = 1:size(addr,2)
        if (addr(1,xx) == to_ind)
            rew_pred_CW_ind(to_ind,addr(2,xx)) = first_conn_ind + xx - 1;
        end;
    end;
end;

%
% Load the DA weights
%

% Find the fascicles.
for xx = 1:size(Conn_name,1)
    the_conn_name = sscanf(Conn_name(xx,:),'%s');
    if (strcmp(the_conn_name, 'TonicDAtoDACell'))
        TonicDAtoDACell_ind = xx;    
    elseif (strcmp(the_conn_name, 'RewCuetoDACell'))
        RewCuetoDACell_ind = xx;
    elseif (strcmp(the_conn_name, 'RewPredtoDACell'))
        RewPredtoDACell_ind = xx;
    end;
end;

% Find the weight for the tonic DA to DA fascicle.
first_conn_ind = 1;
for xx = 1:TonicDAtoDACell_ind - 1
    first_conn_ind = first_conn_ind + NC(xx);
end;
DA_tonic_input_wt = CW(first_conn_ind);
 
% Find the weight for the reward cue to DA fascicle.
first_conn_ind = 1;
for xx = 1:RewCuetoDACell_ind - 1
    first_conn_ind = first_conn_ind + NC(xx);
end;
DA_rew_cue_wt = CW(first_conn_ind);
    
% Find the weights for the reward prediction to DA fascicle.
first_conn_ind = 1;
for xx = 1:RewPredtoDACell_ind - 1
    first_conn_ind = first_conn_ind + NC(xx);
end;
last_conn_ind = first_conn_ind + NC(RewPredtoDACell_ind) - 1;
DA_rew_pred_wts = CW(first_conn_ind:last_conn_ind)';


% Prepare learning algorithm.

% Find the indices for the modifiable (i.e., learning) connections.
[a mdc] = find(MOD(CH));  

if ((~isempty(mdc)) && (learning_on == 1))
	% Find the indices for the modifiable connection fascicle indices.
	[a mdf] = find(MOD(Chan_type));
	
	% Get the list (with replications removed) of all indices of units containing
	% modifiable input weights.
	mdu = unique(CA(1,mdc));
end;

modCW = zeros(size(mdc,2),num_iters);
for iter = 1:num_iters
    modCW(:,iter) = CW(mdc)';
end;

%
% Initialize the simulation
%
trial_num = 1;
trial_iter_count = 0;

% Initialize vectors for unit activity.
% NOTE: Spast and S are ordered so that activity is drawn column-wise 
% from the areas.
Spast = init_unit_noise .* rand(1,sum(NU));      
    % random uniform distribution 0 to 1 * noise factor
   
% Initialize output data arrays.
output = zeros(sum(NU),num_iters);

% Initialize the afferent input.
aff_input1 = 0.0;
aff_input2 = 0.0;

%
% Do the simulation loop.
%

% Loop over all iterations...
for iter = 1:num_iters
    
% Increment the trial iteration count.
trial_iter_count = trial_iter_count + 1;

% Increment the trial count if appropriate.
if (trial_iter_count == pre_CS_iters + CS_iters + 1)
    trial_iter_count = 1;
    trial_num = trial_num + 1;
    aff_input1 = 0.0;
end;

% Print out strings telling how many iterations have passed.
if (mod(iter,monitor_interval) == 0)
    disp(['Iteration ',num2str(iter),' (',num2str(iter * iter_duration), ...
        ' ms)']);
end;

% Change the afferent input if appropriate.
if ((trial_num <= num_crit_trials) && (CS_on(trial_num) == 1))
    if (trial_iter_count == pre_CS_iters + 1)
        aff_input1 = 1.0;
    end;
end;

if ((trial_num <= num_crit_trials) && (R_on(trial_num) == 1))
    if (trial_iter_count == pre_CS_iters + R_CS_offset_iters + 1)
        aff_input2 = 1.0;
    elseif (trial_iter_count == pre_CS_iters + R_CS_offset_iters + ...
            R_iters + 1)
        aff_input2 = 0.0;        
    end;
end;

% Remember the past CS (conditioned stimulus) activities.
CS_past_activs = Spast(CS_WMLoUnit_ind:CS_WMHiUnit_ind)';

% Copy the input information into the appropriate input units.
Spast(CS_WMLoUnit_ind) = aff_input1;
Spast(RewardInUnit_ind) = aff_input2;
Spast(TonicDAAffUnit_ind) = tonic_DA_input;

% Extract the (current) CS activations and the CS to reward prediction 
% weights.
CS_activs = Spast(CS_WMLoUnit_ind:CS_WMHiUnit_ind)';
rew_pred_wts = CW(rew_pred_CW_ind);

% For all of the CSs...
for xx = 1:num_CS
    % Reset all "intracellular calcium" counters for the CS to -1 if the
    % corresponding CS was off on the previous time-step and has just
    % turned on.
    if ((CS_past_activs(xx) <= CS_thresh) && (CS_activs(xx) > CS_thresh))
        rew_pred_ic_ct(:,xx) = -min_rew_delay;
    end;
    
    % Increment every "intracellular calcium" counter for the CS if it is
    % currently on.
    if (CS_activs(xx) > CS_thresh)
        rew_pred_ic_ct(:,xx) = rew_pred_ic_ct(:,xx) + 1;
    end;    
end;

% Find the excitatory net input of the reward prediction delay units.
for yy = 1:num_rew_delays
    for xx = 1:num_CS
        if (rew_pred_ic_ct(yy,xx) == yy)
            rew_pred_act_wts(yy,xx) = rew_pred_wts(yy,xx);
        else
            rew_pred_act_wts(yy,xx) = 0;
        end;
    end;
end;
rew_pred_CT = 3;
rew_pred_netin = rew_pred_act_wts * CS_activs;
rew_pred_out = Spast(RewardPredLoUnit:RewardPredHiUnit)';
rew_pred_out = rew_pred_out + (iter_duration / 1000.0) .* ...
    wcunitupdate(rew_pred_out,rew_pred_netin,thresh(rew_pred_CT), ...
    gain(rew_pred_CT),namp(rew_pred_CT),1 ./ outtimeconst(rew_pred_CT), ...
    1 ./ outtimeconst(rew_pred_CT));

% First sum all of the voltage-independent glutamate inputs.  VI(CH) is the 
% filter for voltage-independent inputs.  ~inhib(CH) filters out the 
% inhibitory units.
Avie = zeros(1,sum(NU)) + sum(sparse(double(CA(1,:)),1:sum(NC),VI(CH) .* ... 
    ~inhib(CH) .* CW .* Spast(CA(2,:))),2)';

% Next, sum all of the voltage-independent GABA inputs.  VI(CH) is the 
% filter for voltage-independent inputs.  inhib(CH) filters the inhibitory 
% units.
Avii = zeros(1,sum(NU)) + sum(sparse(double(CA(1,:)),1:sum(NC),VI(CH) .* ... 
    inhib(CH) .* CW .* Spast(CA(2,:))),2)';

% Then sum all the voltage-dependent inputs.
% vd_thresh = 0.3;
% vd_gain = 50;
% % vd_thresh = 50.0;
% Avd = zeros(1,sum(NU)) + sum(sparse(double(CA(1,:)),1:sum(NC),~VI(CH) ...
%     .*sigm(vd_thresh,vd_gain,Avi(CA(1,:))) ...
%     .*CW.*Spast(CA(2,:))),2)';

% Update unit activity (using Euler's approximation for simulating).
% S = Spast + wcunitupdate(Spast,Avi+Avd,thresh(CT),gain(CT),namp(CT), ...
%     outtimeconst(CT),outtimeconst(CT));
S = Spast + (iter_duration / 1000.0) .* ...
    wcunitupdate(Spast,Avie - Avii,thresh(CT),gain(CT),namp(CT), ...
    1 ./ outtimeconst(CT),1 ./ outtimeconst(CT));

% Update the DA cell activity, from the new reward prediction and reward
% cue outputs.
tonic_DA_out = S(TonicDAAffUnit_ind);
rew_cue_out = S(RewardCueEUnit_ind);
rew_cue_thresh = 0.7;
DACell_CT = 4;
DACell_netin = DA_tonic_input_wt * tonic_DA_input + ...
    DA_rew_cue_wt * rew_cue_out * (rew_cue_out > rew_cue_thresh) - ...
    DA_rew_pred_wts' * rew_pred_out;
DACell_out = Spast(DACellUnit_ind);
DACell_out = DACell_out + (iter_duration / 1000.0) .* ...
    wcunitupdate(DACell_out,DACell_netin,thresh(DACell_CT), ...
    gain(DACell_CT),namp(DACell_CT),1 ./ outtimeconst(DACell_CT), ...
    1 ./ outtimeconst(DACell_CT));

% Copy the inputs back to the initial units so the inputs are correctly 
% remembered.
S(CS_WMLoUnit_ind) = aff_input1;
S(RewardInUnit_ind) = aff_input2;
S(TonicDAAffUnit_ind) = tonic_DA_input;

% Overwrite the adaptive critic output activations calculated in the
% default calculation with the proper outputs.
S(RewardPredLoUnit:RewardPredHiUnit) = rew_pred_out;
S(DACellUnit_ind) = DACell_out;

% Do learning if learning is on.
if (learning_on == 1)
        % LTP term active if post- and pre-synaptic activity are high
        % enough and the reward is on...     
        % LTD term active if post-synaptic activity is high enough and the
        % presynaptic activity is over the LTD threshold but under the LTP
        % threshold, and the reward is on...
        % Weight decay term which is always active...   
    CW(mdc) = CW(mdc) ...     
        + learnrate(CH(mdc)) .* ...
        (S(CA(1,mdc)) > LTP_LTD_postsyn_thresh) .* ...
        (S(CA(2,mdc)) > LTP_presyn_thresh) .* ...
        (S(DACellUnit_ind) > LTP_LTD_rew_thresh) ...
        - learnrate(CH(mdc)) .* LTD_decay_coeff .* ...
        (S(CA(1,mdc)) > LTP_LTD_postsyn_thresh) .* ...
        (S(CA(2,mdc)) > LTD_presyn_thresh) .* ...
        (S(CA(2,mdc)) < LTP_presyn_thresh) .* ...
        (S(DACellUnit_ind) > LTP_LTD_rew_thresh);
 %       - wt_decay_rate; % should multiply by CW(mdc)...
     
    % Clip any weights that have fallen below 0 at 0.
    CW(mdc) = CW(mdc) .* (CW(mdc) >= 0);
      
    % Clip any weights that have risen over the max.
    hi_wts = find(CW(mdc) > maxwt(CH(mdc)));
    CW(mdc(hi_wts)) = maxwt(CH(mdc(hi_wts)));
 
    modCW(:,iter) = CW(mdc)';
    
    % Do learning for the reward prediction cells.
    rew_pred_wts = rew_pred_wts ...
        + rew_pred_learn_rate .* ...
        (rew_pred_act_wts > 0) .* ...
        (S(DACellUnit_ind) > LTP_LTD_rew_thresh);
    
    % Clip any weights that have fallen below 0 at 0.
    rew_pred_wts = rew_pred_wts .* (rew_pred_wts >= 0);
      
    % Clip any weights that have risen over the max.
    hi_wts = find(rew_pred_wts > rew_pred_maxwt);
    rew_pred_wts(hi_wts) = rew_pred_maxwt;
    
    % Copy the weights to CW.
    CW(rew_pred_CW_ind) = rew_pred_wts;
end;

% Save unit activities for later display.
output(:,iter) = S';

% Remember the previous activations.
Spast = S;

% Display the current iteration.
subplot(2,3,1);
colormap(gray);
area_ind = CS_WM_ind; 
disp_area_act; 
xlabel('CS WM');
set(gca,'XTick',[]);
set(gca,'YTick',[]);
    
subplot(2,3,2);
area_ind = RewardCueE_ind; 
disp_area_act; 
xlabel('Reward Cue E'); 
axis square;

subplot(2,3,3);
area_ind = DACell_ind; 
disp_area_act; 
xlabel('DA Cell'); 
axis square;

subplot(2,3,4);
area_ind = RewardIn_ind; 
disp_area_act; 
xlabel('Reward In'); 
axis square;

subplot(2,3,5);
area_ind = RewardPred_ind; 
fetch_area_act; 
imagesc(reshape(activs,5,8), [0 1]);
xlabel('Reward Pred'); 
axis square;

pause(0.01);

end;  % iteration


% Save the brain output
save(cogmodel_out_fname,'output','mdc','modCW','num_iters');

% If we're doing learning and are supposed to, save new weights for the 
% model.
if ((~isempty(mdc)) && (learning_on == 1) && (save_trained_wts == 1))    
    save(train_model_fname,'Model_name','Num_areas','Num_conns', ...
        'Area_name','Area_num','Cell_type','Area_xy','Conn_name', ...
        'Conn_fromto','Chan_type','Conn_kind','Conn_xy','Conn_prob', ...
        'Weig_mean','Weig_sigm','Conn_scle','CA','CW','CR','CH', ...
        'CT','NC','NU','unit_offset');
end;