%
% layer_sim.m
%
% Simulate a layer of units.
%
% Last modification: 11/17/06


% Set simulation parameters.
num_iters = 200;          % number of iterations in simulation (maximum)
iter_duration = 100.0;    % duration of each iteration (in ms)
monitor_interval = 100;   % increment of iterations at which we report how 
                          % much time has elapsed
init_unit_noise = 0.001;  % amplitude of initial unit activity noise
load_trained_wts = 0;
learning_on = 0;
save_trained_wts = 0;
LTP_LTD_postsyn_thresh = 0.7;
LTD_presyn_thresh = 0.3;
LTP_presyn_thresh = 0.7;
LTP_LTD_rew_thresh = 0.7;
LTD_decay_coeff = 0.5;
wt_decay_rate = 0.0001;

%
% Initialize the model.
%

% Load simulation run parameters.
load sim_run_params;

% Load model parameters (connectivity and cell/channel-types).
if (load_trained_wts == 1)
    load(train_model_fname);
else
    load(start_model_fname);
end;

% Load physiological parameters (cells and channels).
load(phys_fname);

% Run through all of the areas and extract the appropriate area indices.
for xx = 1:size(Area_name,1)
    the_area_name = sscanf(Area_name(xx,:),'%s');
    if (strcmp(the_area_name, 'AffInput'))
        AffInput_ind = Area_num(xx);
    elseif (strcmp(the_area_name, 'TestLayerE'))
        TestLayer_ind = Area_num(xx); 
    elseif (strcmp(the_area_name, 'LayerReg'))
        LayerReg_ind = Area_num(xx);         
    end;
end;

% Find the afferent input unit index.
lo_bound = 1;
for xx = 1:AffInput_ind - 1
    lo_bound = lo_bound + Area_xy(xx,3);
end;
hi_bound = lo_bound + Area_xy(AffInput_ind,3) - 1;
AffInputUnit_ind = lo_bound;

% Find the layer regulator unit index.
lo_bound = 1;
for xx = 1:LayerReg_ind - 1
    lo_bound = lo_bound + Area_xy(xx,3);
end;
hi_bound = lo_bound + Area_xy(LayerReg_ind,3) - 1;
LayerRegUnit_ind = lo_bound;

% Prepare learning algorithm.

% Find the indices for the modifiable (i.e., learning) connections.
[a mdc] = find(MOD(CH));  

if ((~isempty(mdc)) && (learning_on == 1))
	% Find the indices for the modifiable connection fascicle indices.
	[a mdf] = find(MOD(Chan_type));
	
	% Get the list (with replications removed) of all indices of units containing
	% modifiable input weights.
	mdu = unique(CA(1,mdc));
end;

modCW = zeros(size(mdc,2),num_iters);
for iter = 1:num_iters
    modCW(:,iter) = CW(mdc)';
end;

% Initialize vectors for unit activity.
% NOTE: Spast and S are ordered so that activity is drawn column-wise 
% from the areas.
Spast = init_unit_noise .* rand(1,sum(NU));      
    % random uniform distribution 0 to 1 * noise factor

% Initialize output data arrays.
output = zeros(sum(NU),num_iters);

% Initialize the afferent input.
aff_input = 0.0;

%
% Do the simulation loop.
%

% Loop over all iterations...
for iter = 1:num_iters
    
% Print out strings telling how many iterations have passed.
if (mod(iter,monitor_interval) == 0)
    disp(['Iteration ',num2str(iter),' (',num2str(iter * iter_duration), ...
        ' ms)']);
end;

% Change the afferent input if appropriate.
if (iter == 50)
    aff_input = 1.0;
elseif (iter == 100)
    aff_input = 0.0;
elseif (iter == 150)
    aff_input = 1.0;
end;

% Copy the input information into the appropriate input units.
Spast(AffInputUnit_ind) = aff_input;

% First sum all of the voltage-independent glutamate inputs.  VI(CH) is the 
% filter for voltage-independent inputs.  ~inhib(CH) filters out the 
% inhibitory units.
Avie = zeros(1,sum(NU)) + sum(sparse(double(CA(1,:)),1:sum(NC),VI(CH) .* ... 
    ~inhib(CH) .* CW .* Spast(CA(2,:))),2)';

% Next, sum all of the voltage-independent GABA inputs.  VI(CH) is the 
% filter for voltage-independent inputs.  inhib(CH) filters the inhibitory 
% units.
Avii = zeros(1,sum(NU)) + sum(sparse(double(CA(1,:)),1:sum(NC),VI(CH) .* ... 
    inhib(CH) .* CW .* Spast(CA(2,:))),2)';

% Then sum all the voltage-dependent inputs.
% vd_thresh = 0.3;
% vd_gain = 50;
% % vd_thresh = 50.0;
% Avd = zeros(1,sum(NU)) + sum(sparse(double(CA(1,:)),1:sum(NC),~VI(CH) ...
%     .*sigm(vd_thresh,vd_gain,Avi(CA(1,:))) ...
%     .*CW.*Spast(CA(2,:))),2)';

% Update unit activity (using Euler's approximation for simulating).
% S = Spast + wcunitupdate(Spast,Avi+Avd,thresh(CT),gain(CT),namp(CT), ...
%     outtimeconst(CT),outtimeconst(CT));
S = Spast + (iter_duration / 1000.0) .* ...
    wcunitupdate(Spast,Avie - Avii,thresh(CT),gain(CT),namp(CT), ...
    1 ./ outtimeconst(CT),1 ./ outtimeconst(CT));

% Copy the inputs back to the initial units so the inputs are correctly 
% remembered.
S(AffInputUnit_ind) = aff_input;

% Do learning if learning is on.
if (learning_on == 1)
        % LTP term active if post- and pre-synaptic activity are high
        % enough and the reward is on...     
        % LTD term active if post-synaptic activity is high enough and the
        % presynaptic activity is over the LTD threshold but under the LTP
        % threshold, and the reward is on...
        % Weight decay term which is always active...   
    CW(mdc) = CW(mdc) ...     
        + learnrate(CH(mdc)) .* ...
        (S(CA(1,mdc)) > LTP_LTD_postsyn_thresh) .* ...
        (S(CA(2,mdc)) > LTP_presyn_thresh) .* ...
        (S(RewardUnit_ind) > LTP_LTD_rew_thresh) ...
        - learnrate(CH(mdc)) .* LTD_decay_coeff .* ...
        (S(CA(1,mdc)) > LTP_LTD_postsyn_thresh) .* ...
        (S(CA(2,mdc)) > LTD_presyn_thresh) .* ...
        (S(CA(2,mdc)) < LTP_presyn_thresh) .* ...
        (S(RewardUnit_ind) > LTP_LTD_rew_thresh);
 %       - wt_decay_rate; % should multiply by CW(mdc)...
     
    % Clip any weights that have fallen below 0 at 0.
    CW(mdc) = CW(mdc) .* (CW(mdc) >= 0);
      
    % Clip any weights that have risen over the max.
    hi_wts = find(CW(mdc) > maxwt(CH(mdc)));
    CW(mdc(hi_wts)) = maxwt(CH(mdc(hi_wts)));
 
    modCW(:,iter) = CW(mdc)';
end;

% Save unit activities for later display.
output(:,iter) = S';

% Remember the previous activations.
Spast = S;

% Display the current iteration.
subplot(1,3,1);
colormap(gray);
area_ind = AffInput_ind; 
disp_area_act; 
xlabel('Afferent Input');
set(gca,'XTick',[]);
set(gca,'YTick',[]);
axis square;
    
subplot(1,3,2);
area_ind = TestLayer_ind; 
disp_area_act; 
xlabel('Layer Units'); 
axis square;
   
subplot(1,3,3);
area_ind = LayerReg_ind; 
disp_area_act; 
xlabel('Layer Regulator');
set(gca,'XTick',[]);
set(gca,'YTick',[]);
axis square;

pause(0.01);

end;  % iteration


% Save the brain output
save(cogmodel_out_fname,'output','mdc','modCW','num_iters');

% If we're doing learning and are supposed to, save new weights for the 
% model.
if ((~isempty(mdc)) && (learning_on == 1) && (save_trained_wts == 1))    
    save(train_model_fname,'Model_name','Num_areas','Num_conns', ...
        'Area_name','Area_num','Cell_type','Area_xy','Conn_name', ...
        'Conn_fromto','Chan_type','Conn_kind','Conn_xy','Conn_prob', ...
        'Weig_mean','Weig_sigm','Conn_scle','CA','CW','CR','CH', ...
        'CT','NC','NU','unit_offset');
end;
